"""Module containing the corrected Fluent GUI help strings."""

XML_HELP_PATCH = {
    "flu_meshing_file_start_transcript": "Starts recording input and output in a file. A transcript file contains a complete record of all standard input to and output from Fluent (usually all keyboard and user interface input and all screen output).Start the transcription process with the file/start-transcript command, and end it with the file/stop-transcript command (or by exiting the program)."
}
